from __future__ import annotations

import typing as t

if t.TYPE_CHECKING:
    from PIL.Image import Image

__all__: t.Final[t.List[str]] = ["calculate_size"]


def calculate_size(
    img: Image, autoscale: bool, width: t.Optional[int], height: t.Optional[int]
) -> t.Tuple[int, int]:
    if autoscale is True:
        from PIL import ImageGrab

        height = ImageGrab.grab().height
        width = int(img.width / img.height * height)

    if width is not None:
        height = int(width * img.height / img.width)

    if height is not None:
        if width is None:
            width = int(img.width / img.height * height)

    return width or img.width, height or img.height



import json
import pathlib
import typing as t

PaletteT = t.List[t.Iterable[int]]

__all__: t.List[str] = ["PaletteT", "load_palette"]


def load_palette(path: str) -> PaletteT:
    path = pathlib.Path(path)
    if path.exists() is False:
        raise FileNotFoundError(f"{path} does not exist.")

    if path.is_dir() is True:
        raise IsADirectoryError(f"{path} is a directory.")

    with open(str(path)) as fp:
        palette = json.load(fp)

    if not isinstance(palette, list) and not all(
        isinstance(channel, int) for rgb in palette for channel in rgb
    ):
        raise ValueError("JSON file must contain list of RGB channels.")

    return [tuple(rgb) for rgb in palette]



import typing as t

from palettify.internal import loader, scale
from palettify.internal.convert import *
from palettify.internal.loader import *
from palettify.internal.scale import *

__all__: t.List[str] = [*loader.__all__, *scale.__all__]



import os
import re
from setuptools import setup
from setuptools import find_namespace_packages
import types

from Cython.Build import cythonize

pkg_name = "palettify"

def parse_meta():
    with open(os.path.join(pkg_name, "_about.py")) as fp:
        code = fp.read()

    token_pattern = re.compile(r"^__(?P<key>\w+)?__\s*=\s*(?P<quote>(?:'{3}|\"{3}|'|\"))(?P<value>.*?)(?P=quote)", re.M)

    groups = {}

    for match in token_pattern.finditer(code):
        group = match.groupdict()
        groups[group["key"]] = group["value"]

    return types.SimpleNamespace(**groups)


def long_description():
    with open("README.md") as fp:
        return fp.read()


def parse_requirements_file(path):
    with open(path) as fp:
        dependencies = (d.strip() for d in fp.read().split("\n") if d.strip())
        return [d for d in dependencies if not d.startswith("#")]

meta = parse_meta()

setup(
    name=pkg_name,
    version=meta.version,
    description="Apply color palettes on images.",
    long_description=long_description(),
    long_description_content_type="text/markdown",
    author=meta.author,
    author_email=meta.email,
    url=meta.url,
    packages=find_namespace_packages(include=[pkg_name + "*"]),
    entry_points={"console_scripts": ["palettify = palettify.cli:main"]},
    license=meta.license,
    include_package_data=True,
    zip_safe=False,
    ext_modules=cythonize(f"{pkg_name}/internal/convert.pyx"),
    install_requires=parse_requirements_file("requirements.txt"),
    python_requires=">=3.6.0,<3.11",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]
)


from __future__ import annotations

import logging
import multiprocessing
import pathlib
import threading
import typing as t

from PIL import Image

from palettify.internal import apply_palette, calculate_size, load_palette

if t.TYPE_CHECKING:
    from palettify.internal import PaletteT

_LOGGER: t.Final[logging.Logger] = logging.getLogger(__name__)

__all__: t.List[str] = ["palettify_image"]


def _process_image(
    input_path: str,
    output_path: str,
    palette: t.Optional[PaletteT],
    auto_palette: t.Optional[int],
    autoscale: bool,
    width: t.Optional[int],
    height: t.Optional[int],
    threads: int,
) -> None:
    img = Image.open(input_path)
    width, height = calculate_size(
        img=img, autoscale=autoscale, width=width, height=height
    )

    size = (width, height)
    if size != img.size:
        img = img.resize(size)
        _LOGGER.info("Resized image to %sx%s", width, height)

    if auto_palette is not None:
        _LOGGER.info(
            "Creating color palette using %s most dominant colors.", auto_palette
        )
        palette = [
            rgb
            for _, rgb in sorted(
                img.getcolors(16777216), key=lambda t: t[0], reverse=True
            )[:auto_palette]
        ]

    pixels = img.load()
    pixel_step = width // threads

    threads_started: t.List[threading.Thread] = []

    _LOGGER.info("Starting %s threads for %s.", threads, input_path)
    for x in range(pixel_step, width + pixel_step, pixel_step):
        t = threading.Thread(
            target=apply_palette, args=(pixels, palette, x - pixel_step, x, height)
        )
        t.daemon = True
        t.start()
        threads_started.append(t)

    for t in threads_started:
        t.join()

    _LOGGER.info("Saving image to %s.", output_path)
    img.save(output_path)


def _path_duplicate(path: pathlib.Path) -> str:
    return path.stem + "-palettified" + path.suffix


def palettify_image(
    paths: t.Iterable[str],
    palette: str,
    *,
    autoscale: bool = False,
    output: t.Optional[str] = None,
    overwrite: bool = False,
    width: t.Optional[int] = None,
    height: t.Optional[int] = None,
    threads: int = 10,
    verbose: bool = False,
) -> None:
    if verbose is True:
        logging.basicConfig(format="%(message)s", level="INFO")

    auto_palette: t.Optional[int] = None
    _LOGGER.info("Loading color palette.")
    if not palette.isdigit():
        palette = load_palette(palette)

    else:
        auto_palette = int(palette)
        _LOGGER.info(
            "Palette detection is enabled. Color palette will be created using the %s most dominant colors.",
            auto_palette,
        )

    processes_started: t.List[multiprocessing.Process] = []

    for input_path in paths:
        _LOGGER.info("Processing %s.", input_path)
        input_path = pathlib.Path(input_path)
        if not input_path.exists():
            raise FileNotFoundError(f"{input_path} does not exist.")

        if overwrite is True:
            _LOGGER.info("Overwrite is enabled, using input path as output path.")
            output_path = input_path

        elif output is not None:
            output_path = pathlib.Path(output)
            _LOGGER.info("Processing %s.", output_path)
            if output_path.exists() and output_path.is_dir():
                output_path /= _path_duplicate(input_path)

            elif len(paths) > 1:
                raise NotADirectoryError(
                    "A directory needs to be passed for multiple input paths."
                )

        else:
            output_path = input_path.parent / _path_duplicate(input_path)
            _LOGGER.info("No output path is given, defaulting to %s.", output_path)

        _LOGGER.info("Starting process for %s.", input_path)
        p = multiprocessing.Process(
            target=_process_image,
            args=(
                str(input_path),
                str(output_path),
                palette,
                auto_palette,
                autoscale,
                width,
                height,
                threads,
            ),
        )
        p.daemon = True
        p.start()
        processes_started.append(p)

    for p in processes_started:
        p.join()



import argparse
import sys

from palettify import palettify_image


def excepthook(_, error: BaseException, __) -> None:
    print(f"\33[31m{error}\33[0m")


def main():
    sys.excepthook = excepthook
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "paths", nargs="+", help="The paths to the images to apply the palette to."
    )
    parser.add_argument(
        "-p",
        "--palette",
        required=True,
        help="The JSON file to load the palette from or the amount of the most dominant colors in the original image to use for the palette.",
    )
    parser.add_argument(
        "-o",
        "--output",
        required=False,
        help="Path to the file or directory to write the image(s) to.",
    )
    parser.add_argument(
        "-O",
        "--overwrite",
        required=False,
        action="store_true",
        help="Whether to overwrite the original images.",
    )
    parser.add_argument(
        "-a",
        "--autoscale",
        required=False,
        action="store_true",
        help="Whether to resize the image to the screen resolution.",
    )
    parser.add_argument(
        "-w", "--width", required=False, type=int, help="The width of the image."
    )
    parser.add_argument(
        "-H", "--height", required=False, type=int, help="The height of the image."
    )
    parser.add_argument(
        "-t",
        "--threads",
        required=False,
        type=int,
        default=10,
        help="The number of threads to start.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        required=False,
        action="store_true",
        help="Explain what is being done.",
    )
    args = parser.parse_args()

    palettify_image(**args.__dict__)
